use crate::app_config::AppConfig;
use crate::constants;
use crate::controllers::HttpResult;
use crate::converters::finances_manager_converter::{
    FinancesManagerAccount, FinancesManagerFile, FinancesManagerMovement,
};
use crate::extractors::auth_extractor::AuthExtractor;
use crate::extractors::file_extractor::FileExtractor;
use crate::models::accounts::AccountType;
use crate::services::accounts_service::UpdateAccountQuery;
use crate::services::movements_service::UpdateMovementQuery;
use crate::services::{accounts_service, files_service, movements_service};
use crate::utils::time_utils::{format_date, time};
use actix_files::NamedFile;
use actix_web::{HttpRequest, HttpResponse};
use serde::Serialize;
use std::fs::File;
use std::io::Write;
use zip::ZipWriter;
use zip::write::SimpleFileOptions;

/// Generate export filename
fn export_filename(ext: &str) -> String {
    format!(
        "export_{}.{ext}",
        format_date(time() as i64).unwrap().replace('/', "-")
    )
}

/// Import data from a [FinancesManager](https://gitlab.com/pierre42100/cpp-financesmanager) file
pub async fn finances_manager_import(auth: AuthExtractor, file: FileExtractor) -> HttpResult {
    let file = FinancesManagerFile::parse(&String::from_utf8_lossy(&file.buff))?;

    // Create each account & push the movements independently
    for file_account in file.accounts {
        let account = accounts_service::create(
            auth.user_id(),
            &UpdateAccountQuery {
                name: file_account.name,
                r#type: AccountType::Cash,
            },
        )
        .await?;

        for file_movement in file_account.movements {
            movements_service::create(&UpdateMovementQuery {
                account_id: account.id(),
                time: file_movement.time,
                label: file_movement.label,
                file_id: None,
                amount: file_movement.amount,
                checked: false,
            })
            .await?;
        }
    }

    Ok(HttpResponse::Accepted().finish())
}

/// Export data to a [FinancesManager](https://gitlab.com/pierre42100/cpp-financesmanager) file
pub async fn finances_manager_export(auth: AuthExtractor) -> HttpResult {
    let accounts = accounts_service::get_list_user(auth.user_id()).await?;

    let mut out = FinancesManagerFile { accounts: vec![] };

    for account in accounts {
        let movements = movements_service::get_list_account(account.id()).await?;
        let mut file_account = FinancesManagerAccount {
            name: account.name,
            movements: Vec::with_capacity(movements.len()),
        };

        for movement in movements {
            file_account.movements.push(FinancesManagerMovement {
                label: movement.label,
                time: movement.time as u64,
                amount: movement.amount,
            });
        }

        out.accounts.push(file_account);
    }

    Ok(HttpResponse::Ok()
        .insert_header((
            "Content-Disposition",
            format!("attachment; filename={}", export_filename("finance")),
        ))
        .body(out.encode()))
}

/// Add JSON file to ZIP
fn zip_json<E: Serialize>(
    zip: &mut ZipWriter<File>,
    path: &str,
    content: &E,
) -> anyhow::Result<()> {
    let file_encoded = serde_json::to_string(&content)?;

    let options = SimpleFileOptions::default()
        .compression_method(zip::CompressionMethod::Deflated)
        .unix_permissions(0o750);

    zip.start_file(path, options)?;
    zip.write_all(file_encoded.as_bytes())?;
    Ok(())
}

/// Export all user data as ZIP
pub async fn zip_export(req: HttpRequest, auth: AuthExtractor) -> HttpResult {
    let dest_dir = tempfile::tempdir_in(&AppConfig::get().temp_dir)?;
    let zip_path = dest_dir.path().join("export.zip");

    let file = File::create(&zip_path)?;
    let mut zip = ZipWriter::new(file);

    // Process all JSON documents
    zip_json(
        &mut zip,
        constants::zip_export::ACCOUNTS_FILE,
        &accounts_service::get_list_user(auth.user_id()).await?,
    )?;
    zip_json(
        &mut zip,
        constants::zip_export::MOVEMENTS_FILE,
        &movements_service::get_all_movements_user(auth.user_id()).await?,
    )?;
    let files_list = files_service::get_all_files_user(auth.user_id()).await?;
    zip_json(&mut zip, constants::zip_export::FILES_FILE, &files_list)?;
    // TODO : inbox

    // Process all files
    for file in files_list {
        let buff = files_service::get_file_content(&file).await?;

        let options = SimpleFileOptions::default()
            .compression_method(zip::CompressionMethod::Deflated)
            .unix_permissions(0o750);

        zip.start_file(
            format!("{}{}", constants::zip_export::FILES_DIR, file.sha512),
            options,
        )?;
        zip.write_all(&buff)?;
    }

    // Finalize ZIP and return response
    zip.finish()?;
    let file = File::open(zip_path)?;

    let file = NamedFile::from_file(file, export_filename("zip"))?;

    Ok(file.into_response(&req))
}
